#include "pfodBufferedClient.h"
/**
  (c)2015 Forward Computing and Control Pty. Ltd.
  This code may be freely used for both private and commerical use.
  Provide this copyright is maintained.
*/

// uncomment this next line and call setDebugStream(&Serial); to enable debug out
//#define DEBUG

void pfodBufferedClient::setDebugStream(Print* out) {
  debugOut = out;
}

pfodBufferedClient::pfodBufferedClient() {
  client = NULL;
  debugOut = NULL;
  // bufferSize for now always 1456
  sendDelayTime = DEFAULT_SEND_DELAY_TIME;
  sendBufferIdx = 0;
}

pfodBufferedClient* pfodBufferedClient::connect(Client* _client) {
#ifdef DEBUG
  if (debugOut) {
    debugOut->println("called connect");
  }
#endif // DEBUG		
  client = _client;
  sendBufferIdx = 0;
  return this;
}

void pfodBufferedClient::stop() {
  sendAfterDelay();
  if (!client) {
    return;
  }
  client->stop();
  client = 0;
}


uint8_t pfodBufferedClient::connected() {
  sendAfterDelay();
  if (!client) {
    return 0;
  }
  return  client->connected();
}

size_t pfodBufferedClient::write(const uint8_t *buf, size_t size) {
  sendAfterDelay();
  if (!client) {
    return 0;
  }
  for (size_t i = 0; i < size; i++) {
    _write(buf[i]); // may block if buffer fills and client starts sending packet
  }
  return size;
}

size_t pfodBufferedClient::write(uint8_t c) {
  sendAfterDelay();
  if (!client) {
    return 0;
  }
  return _write(c);
}

size_t pfodBufferedClient::_write(uint8_t c) {
  if (!client) {
    return 0;
  }
  size_t rtn = 1;
  sendTimerStart = millis();
  sendBuffer[sendBufferIdx++] = c;
  //push UART data to connected client, if buffer full OR send timer times out after data stops coming
  if (sendBufferIdx == DEFAULT_SEND_BUFFER_SIZE) {
    unsigned long startWrite = micros();
#ifdef DEBUG
//    if (debugOut != NULL) {
//      debugOut->print("buffer full write "); debugOut->print(sendBufferIdx); debugOut->println(" bytes to client");
//      debugOut->println(millis());
//    }
#endif // DEBUG    
    if (client && client->connected()) {
      // buffer full
      // returns ((size_t)-1) if cannot write in 5 sec
      if (debugOut != NULL) {
        debugOut->print(micros()); debugOut->print(" ");
      }
      rtn = client->write((const uint8_t *)sendBuffer, sendBufferIdx); // this call may block if last packet not ACKed yet  
      // This call takes about 350uS
      
#ifdef DEBUG
      if (debugOut != NULL) {
        debugOut->println(micros() - startWrite); //debugOut->println(" after write");
      }
#endif // DEBUG    
    } else {
#ifdef DEBUG
      //    if (debugOut != NULL) {
      //      debugOut->print("client not connected do nothing ");
      //      debugOut->println(millis());
      //    }
#endif // DEBUG
      // just throw this data away
    }
    sendBufferIdx = 0;
  }
  delay(0); // yield
  return rtn;
}

void pfodBufferedClient::sendAfterDelay() {
  if ((!client) || (!sendBufferIdx)) {  // common cases
    return;
  }
  if (((millis() - sendTimerStart) > sendDelayTime)) {
    if (client->connected()) {
      unsigned long startWrite = millis();
#ifdef DEBUG
      if (debugOut != NULL) {
        debugOut->print("sendAfterDelay() "); debugOut->print(sendBufferIdx); debugOut->println(" bytes ");
        debugOut->print(millis());debugOut->print(" ");
      }
#endif // DEBUG    
      client->write((const uint8_t *)sendBuffer, sendBufferIdx);
      // client.flush();  // don't flush does not affect write but does clear inbound buffer :-(
#ifdef DEBUG
      if (debugOut != NULL) {
        debugOut->println(millis() - startWrite);
      }
#endif // DEBUG    
      sendBufferIdx = 0;
      //    }
    } else {
      // just throw this data away
      sendBufferIdx = 0;
    }
  }
  delay(0); // yield
}

// force send the buffer block if necessary
void pfodBufferedClient::forceSend() {
  if ((!client) || (!sendBufferIdx)) {  // common cases
    return;
  }
  if (client->connected()) {
#ifdef DEBUG
    if (debugOut != NULL) {
      debugOut->println("forceSend");
    }
#endif // DEBUG    
    client->write((const uint8_t *)sendBuffer, sendBufferIdx); // this call may block if last packet not ACKed yet
    sendBufferIdx = 0;
  }
  delay(0);
}


// expect available to ALWAYS called before read() so update timer here
int pfodBufferedClient::available() {
  sendAfterDelay();
  if (!client) {
    return 0;
  }
  if (!client->connected()) {
    return 0;
  }
  return client->available();
}

int pfodBufferedClient::read() {
  sendAfterDelay();
  if (!client) {
    return -1;
  }
  int c = client->read();
  delay(0);
  return c;
}

int pfodBufferedClient::peek() {
  sendAfterDelay();
  if (!client) {
    return -1;
  }
  return client->peek();
}

/**
  Forces send of any buffered data now. May block if last packet no ACKed yet
*/
void pfodBufferedClient::flush() {
  forceSend();
  // do not call flush!! for WiFiClient flush() discards unread incoming data;
}
